@@ -6,15 +6,38 @@
  */
 #include "waylandimserverv2.h"
 #include <sys/mman.h>
+#include <sys/types.h>
+#include <algorithm>
+#include <cstdint>
+#include <cstring>
 #include <ctime>
-#include "fcitx-utils/keysymgen.h"
+#include <memory>
+#include <string>
+#include <tuple>
+#include <utility>
+#include <wayland-client-core.h>
+#include <wayland-client-protocol.h>
+#include <xkbcommon/xkbcommon.h>
+#include "fcitx-utils/capabilityflags.h"
+#include "fcitx-utils/event.h"
+#include "fcitx-utils/eventloopinterface.h"
+#include "fcitx-utils/key.h"
+#include "fcitx-utils/keysym.h"
+#include "fcitx-utils/macros.h"
+#include "fcitx-utils/textformatflags.h"
 #include "fcitx-utils/unixfd.h"
 #include "fcitx-utils/utf8.h"
+#include "fcitx/event.h"
 #include "fcitx/inputcontext.h"
+#include "fcitx/instance.h"
 #include "virtualinputcontext.h"
 #include "wayland-text-input-unstable-v3-client-protocol.h"
+#include "wayland_public.h"
 #include "waylandim.h"
+#include "waylandimserverbase.h"
 #include "wl_seat.h"
+#include "zwp_input_method_manager_v2.h"
+#include "zwp_virtual_keyboard_manager_v1.h"
 
 namespace fcitx {
 
@@ -97,7 +120,7 @@ void WaylandIMServerV2::refreshSeat() {
     }
     auto seats = display_->getGlobals<wayland::WlSeat>();
     for (const auto &seat : seats) {
-        if (icMap_.count(seat.get())) {
+        if (icMap_.contains(seat.get())) {
             continue;
         }
         auto *ic =
@@ -125,6 +148,10 @@ WaylandIMInputContextV2::WaylandIMInputContextV2(
     : VirtualInputContextGlue(inputContextManager), server_(server),
       seat_(std::move(seat)),
       ic_(server->inputMethodManagerV2()->getInputMethod(seat_.get())) {
+    if (server_->parent()->persistentVirtualKeyboard()) {
+        vk_.reset(server_->virtualKeyboardManagerV1()->createVirtualKeyboard(
+            seat_.get()));
+    }
     server->add(this, seat_.get());
     ic_->surroundingText().connect(
         [this](const char *text, uint32_t cursor, uint32_t anchor) {
@@ -165,16 +192,20 @@ WaylandIMInputContextV2::WaylandIMInputContextV2(
                 }
                 focusOutWrapper();
             }
-            vk_.reset();
-            vkReady_ = false;
+            if (!server_->parent()->persistentVirtualKeyboard()) {
+                vk_.reset();
+                vkReady_ = false;
+            }
         }
         if (pendingActivate_) {
             pendingActivate_ = false;
-            vk_.reset();
-            vkReady_ = false;
-            vk_.reset(
-                server_->virtualKeyboardManagerV1()->createVirtualKeyboard(
-                    seat_.get()));
+            if (!server_->parent()->persistentVirtualKeyboard()) {
+                vk_.reset();
+                vkReady_ = false;
+                vk_.reset(
+                    server_->virtualKeyboardManagerV1()->createVirtualKeyboard(
+                        seat_.get()));
+            }
             // There can be only one grab. Always release old grab first.
             // It is possible when switching between two client, there will be
             // two activate. In that case we will have already one grab. The
@@ -425,8 +456,10 @@ void WaylandIMInputContextV2::keymapCallback(uint32_t format, int32_t fd,
         1 << xkb_keymap_mod_get_index(server_->keymap_.get(), "Mod5");
 
     if (keymapChanged || !vkReady_) {
-        vk_->keymap(format, scopeFD.fd(), size);
-        vkReady_ = true;
+        if (vk_) {
+            vk_->keymap(format, scopeFD.fd(), size);
+            vkReady_ = true;
+        }
     }
 
     server_->parent_->wayland()->call<IWaylandModule::reloadXkbOption>();
@@ -464,7 +497,8 @@ void WaylandIMInputContextV2::keyCallback(uint32_t serial, uint32_t time,
             repeatSym_ = event.rawKey().sym();
             // Let's trick the key event system by fake our first.
             // Remove 100 from the initial interval.
-            timeEvent_->setNextInterval(repeatDelay() * 1000 - repeatHackDelay);
+            timeEvent_->setNextInterval((repeatDelay() * 1000) -
+                                        repeatHackDelay);
             timeEvent_->setOneShot();
         }
     }
@@ -486,7 +520,7 @@ void WaylandIMInputContextV2::keyCallback(uint32_t serial, uint32_t time,
         WAYLANDIM_DEBUG() << "Engine handling speed can not keep up with key "
                              "repetition rate.";
         timeEvent_->setNextInterval(
-            std::clamp(0, repeatDelay() * 1000 - repeatHackDelay, 1000));
+            std::clamp((repeatDelay() * 1000) - repeatHackDelay, 0, 1000));
     }
 }
 void WaylandIMInputContextV2::modifiersCallback(uint32_t /*serial*/,
@@ -505,8 +539,9 @@ void WaylandIMInputContextV2::modifiersCallback(uint32_t /*serial*/,
     server_->instance()->updateXkbStateMask(
         server_->group()->display(), mods_depressed, mods_latched, mods_locked);
     mask = xkb_state_serialize_mods(
-        server_->state_.get(), static_cast<xkb_state_component>(
-                                   XKB_STATE_DEPRESSED | XKB_STATE_LATCHED));
+        server_->state_.get(),
+        static_cast<xkb_state_component>(XKB_STATE_MODS_DEPRESSED |
+                                         XKB_STATE_MODS_LATCHED));
 
     server_->modifiers_ = 0;
     if (mask & server_->stateMask_.shift_mask) {
